require "rubygems"
require "redis"

class RedisSafeQueue

  # redis must process all commands within 10s
  TX_COMMIT_TIMEOUT = 10

  attr_accessor :timeout, :queue_id, :prefix, :redis, :iterations

  def initialize(opts)
    @redis = opts.fetch(:redis)
    @prefix = opts[:redis_prefix] || :redis_safe_queue
    @queue_id = opts.fetch(:queue_id)
    @timeout = opts.fetch(:timeout).to_i
    @iterations = 0
  end

  def push(job_data)
    job_id = unique_id
    @redis.set(key(job_id, :job), job_data)
    @redis.sadd(key, job_id)
  end

  def work(max_jobs = nil)
    loop do
      break if max_jobs == 0
      if job_id = start_tx
        job = get_job(job_id)
        yield(job)
        commit_tx(job_id)
        max_jobs -= 1 if max_jobs
      else
        break
      end
    end
  end

  def finish
    sleep 1 while size > 0

    lock = @redis.setnx(key(:finished), 1)
    yield if lock

    @redis.expire(key, @timeout)
    @redis.expire(key(:finished), @timeout)
  end

  def size
    @redis.scard(key)
  end

private

  def start_tx
    job_id = nil

    loop do
      @iterations +=1

      candidate = @redis.srandmember(key)
      return unless candidate

      if lock = @redis.setnx(key(candidate, :lock), Time.now.to_i)
        @redis.expire(key(candidate, :lock), @timeout)
        return candidate
      else
        lock_time = @redis.get(key(candidate, :lock)).to_i

        if (Time.now.to_i - lock_time) > @timeout
          @redis.del(key(candidate, :lock))
        end
      end
    end
  end

  def commit_tx(job_id)
    @redis.pipelined do
      @redis.expire(key(job_id, :lock), TX_COMMIT_TIMEOUT)
      @redis.srem(key, job_id)
      @redis.del(key(job_id, :job))
    end
  end

  def unique_id
    rand(8**32).to_s(36)
  end

  def get_job(job_id)
    @redis.get(key(job_id, :job))
  end

  def key(*append)
    [@prefix, @queue_id, append].flatten.compact * ":"
  end

end
